@@ -259,7 +259,7 @@ const char* mysql_fix_privilege_tables={
 "EXECUTE stmt;\n "
 "DROP PREPARE stmt;\n "
 "\n "
-"SET @str=IF(@have_ndbinfo,'CREATE TABLE `ndbinfo`.`ndb$transporters` (`node_id` INT UNSIGNED,`remote_node_id` INT UNSIGNED,`connection_status` INT UNSIGNED) COMMENT=\"transporter status\" ENGINE=NDBINFO','SET @dummy = 0');\n "
+"SET @str=IF(@have_ndbinfo,'CREATE TABLE `ndbinfo`.`ndb$transporters` (`node_id` INT UNSIGNED,`remote_node_id` INT UNSIGNED,`connection_status` INT UNSIGNED,`remote_address` VARCHAR(512),`bytes_sent` BIGINT UNSIGNED,`bytes_received` BIGINT UNSIGNED) COMMENT=\"transporter status\" ENGINE=NDBINFO','SET @dummy = 0');\n "
 "PREPARE stmt FROM @str;\n "
 "EXECUTE stmt;\n "
 "DROP PREPARE stmt;\n "
@@ -360,7 +360,7 @@ const char* mysql_fix_privilege_tables={
 "DROP PREPARE stmt;\n "
 "\n "
 "# ndbinfo.transporters\n "
-"SET @str=IF(@have_ndbinfo,'CREATE OR REPLACE DEFINER=`root@localhost` SQL SECURITY INVOKER VIEW `ndbinfo`.`transporters` AS SELECT node_id, remote_node_id,  CASE connection_status  WHEN 0 THEN \"CONNECTED\"  WHEN 1 THEN \"CONNECTING\"  WHEN 2 THEN \"DISCONNECTED\"  WHEN 3 THEN \"DISCONNECTING\"  ELSE NULL  END AS status FROM ndbinfo.ndb$transporters','SET @dummy = 0');\n "
+"SET @str=IF(@have_ndbinfo,'CREATE OR REPLACE DEFINER=`root@localhost` SQL SECURITY INVOKER VIEW `ndbinfo`.`transporters` AS SELECT node_id, remote_node_id,  CASE connection_status  WHEN 0 THEN \"CONNECTED\"  WHEN 1 THEN \"CONNECTING\"  WHEN 2 THEN \"DISCONNECTED\"  WHEN 3 THEN \"DISCONNECTING\"  ELSE NULL  END AS status,  remote_address, bytes_sent, bytes_received FROM `ndbinfo`.`ndb$transporters`','SET @dummy = 0');\n "
 "PREPARE stmt FROM @str;\n "
 "EXECUTE stmt;\n "
 "DROP PREPARE stmt;\n "
