@@ -1,5 +1,5 @@
 /*
-   Copyright (c) 2003, 2010, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2003, 2011, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -28,6 +28,8 @@
 #include <signaldata/CopyGCIReq.hpp>
 #include <blocks/mutexes.hpp>
 #include <signaldata/LCP.hpp>
+#include <NdbSeqLock.hpp>
+#include <CountingSemaphore.hpp>
 
 #ifdef DBDIH_C
 
@@ -68,6 +70,12 @@
 // Error Codes for Add Table
 // --------------------------------------
 #define ZREPLERROR1 306
+#define ZREPLERROR2 307
+
+// --------------------------------------
+// Other DIH error codes
+// --------------------------------------
+#define ZLONG_MESSAGE_ERROR 312
 
 // --------------------------------------
 // Crash Codes
@@ -98,12 +106,33 @@
 /*#########*/
 /* SIZES   */
 /*#########*/
-#define ZPAGEREC 100
+/*
+ * Pages are used for flushing table definitions during LCP,
+ * and for other operations such as metadata changes etc
+ * 
+ */
+#define MAX_CONCURRENT_LCP_TAB_DEF_FLUSHES 4
+#define MAX_CONCURRENT_DIH_TAB_DEF_OPS (MAX_CONCURRENT_LCP_TAB_DEF_FLUSHES + 2)
+#define ZPAGEREC (MAX_CONCURRENT_DIH_TAB_DEF_OPS * PACK_TABLE_PAGES)
 #define ZCREATE_REPLICA_FILE_SIZE 4
 #define ZPROXY_MASTER_FILE_SIZE 10
 #define ZPROXY_FILE_SIZE 10
 #endif
 
+/*
+ * Pack table into pages.
+ * See use of writePageWord() in
+ * packTableIntoPagesLab() and helper
+ * functions to determine the constants
+ * below.
+ */
+#define MAX_CRASHED_REPLICAS 8
+#define PACK_REPLICAS_WORDS (4 + 4 * MAX_LCP_STORED + 2 * MAX_CRASHED_REPLICAS)
+#define PACK_FRAGMENT_WORDS (6 + 2 * MAX_REPLICAS * PACK_REPLICAS_WORDS)
+#define PACK_TABLE_WORDS (10 + MAX_NDB_PARTITIONS * PACK_FRAGMENT_WORDS)
+#define PACK_TABLE_PAGE_WORDS (2048 - 32)
+#define PACK_TABLE_PAGES ((PACK_TABLE_WORDS + PACK_TABLE_PAGE_WORDS - 1) / PACK_TABLE_PAGE_WORDS)
+
 class Dbdih: public SimulatedBlock {
 #ifdef ERROR_INSERT
   typedef void (Dbdih::* SendFunction)(Signal*, Uint32, Uint32);
@@ -120,7 +149,7 @@ public:
    * ONGOING */
   struct ApiConnectRecord {
     Uint64 apiGci;
-    Uint32 nextApi;
+    Uint32 senderData;
   };
   typedef Ptr<ApiConnectRecord> ApiConnectRecordPtr;
 
@@ -137,14 +166,33 @@ public:
     enum ConnectState {
       INUSE = 0,
       FREE = 1,
-      STARTED = 2
+      STARTED = 2,
+      ALTER_TABLE = 3,
+      ALTER_TABLE_ABORT = 4, // "local" abort
+      ALTER_TABLE_REVERT = 5,
+      GET_TABINFO = 6
+    };
+    union {
+      Uint32 nodes[MAX_REPLICAS];
+      struct {
+        Uint32 m_changeMask;
+        Uint32 m_totalfragments;
+        Uint32 m_org_totalfragments;
+        Uint32 m_new_map_ptr_i;
+      } m_alter;
+      struct {
+        Uint32 m_map_ptr_i;
+      } m_create;
+      struct {
+        Uint32 m_requestInfo;
+      } m_get_tabinfo;
     };
-    Uint32 nodes[MAX_REPLICAS];
     ConnectState connectState;
-    Uint32 nfConnect;
+    Uint32 nextPool;
     Uint32 table;
     Uint32 userpointer;
     BlockReference userblockref;
+    Callback m_callback;
   };
   typedef Ptr<ConnectRecord> ConnectRecordPtr;
 
@@ -163,7 +211,6 @@ public:
     Uint16 logNodeId[MAX_LOG_EXEC];
     Uint32 createLcpId;
 
-    bool hotSpareUse;
     Uint32 replicaRec;
     Uint16 dataNodeId;
     Uint16 lcpNo;
@@ -253,6 +300,9 @@ public:
     Uint32 nextReplicaNode;
     Uint32 nodeCount;
     Uint32 activeTakeOver; // Which node...
+    Uint32 m_next_log_part;
+    Uint32 nodegroupIndex;
+    Uint32 m_ref_count;
   };
   typedef Ptr<NodeGroupRecord> NodeGroupRecordPtr;
   /**/
@@ -419,7 +469,17 @@ public:
    * TO LOCATE A FRAGMENT AND TO TRANSLATE A KEY OF A TUPLE TO THE FRAGMENT IT
    * BELONGS
    */
-  struct TabRecord {
+  struct TabRecord
+  {
+    TabRecord() { }
+
+    /**
+     * rw-lock that protects multiple parallel DIGETNODES (readers) from
+     *   updates to fragmenation changes (e.g CREATE_FRAGREQ)...
+     *   search for DIH_TAB_WRITE_LOCK
+     */
+    NdbSeqLock m_lock;
+
     /**
      * State for copying table description into pages
      */
@@ -434,7 +494,10 @@ public:
       CS_COPY_NODE_STATE,
       CS_ADD_TABLE_MASTER,
       CS_ADD_TABLE_SLAVE,
-      CS_INVALIDATE_NODE_LCP
+      CS_INVALIDATE_NODE_LCP,
+      CS_ALTER_TABLE,
+      CS_COPY_TO_SAVE
+      ,CS_GET_TABINFO
     };
     /**
      * State for copying pages to disk
@@ -442,11 +505,13 @@ public:
     enum UpdateState {
       US_IDLE,
       US_LOCAL_CHECKPOINT,
+      US_LOCAL_CHECKPOINT_QUEUED,
       US_REMOVE_NODE,
       US_COPY_TAB_REQ,
       US_ADD_TABLE_MASTER,
       US_ADD_TABLE_SLAVE,
-      US_INVALIDATE_NODE_LCP
+      US_INVALIDATE_NODE_LCP,
+      US_CALLBACK
     };
     enum TabLcpStatus {
       TLS_ACTIVE = 1,
@@ -463,7 +528,8 @@ public:
       LINEAR_HASH = 0,
       NOTDEFINED = 1,
       NORMAL_HASH = 2,
-      USER_DEFINED = 3
+      USER_DEFINED = 3,
+      HASH_MAP = 4
     };
     enum Storage {
       ST_NOLOGGING = 0,         // Table is not logged, but survives SR
@@ -477,39 +543,44 @@ public:
     Method method;
     Storage tabStorage;
 
-    Uint32 pageRef[8];
+    Uint32 pageRef[PACK_TABLE_PAGES]; // TODO: makedynamic
 //-----------------------------------------------------------------------------
 // Each entry in this array contains a reference to 16 fragment records in a
 // row. Thus finding the correct record is very quick provided the fragment id.
 //-----------------------------------------------------------------------------
-    Uint32 startFid[MAX_NDB_NODES];
+    Uint32 startFid[(MAX_NDB_PARTITIONS - 1) / NO_OF_FRAGS_PER_CHUNK + 1];
 
     Uint32 tabFile[2];
     Uint32 connectrec;                                    
-    Uint32 hashpointer;
-    Uint32 mask;
+    union {
+      Uint32 hashpointer;
+      Uint32 m_new_map_ptr_i;
+    };
+    union {
+      Uint32 mask;
+      Uint32 m_map_ptr_i;
+    };
     Uint32 noOfWords;
     Uint32 schemaVersion;
     Uint32 tabRemoveNode;
     Uint32 totalfragments;
     Uint32 noOfFragChunks;
+    Uint32 m_scan_count[2];
+    Uint32 m_scan_reorg_flag;
     Uint32 tabErrorCode;
     struct {
       Uint32 tabUserRef;
       Uint32 tabUserPtr;
     } m_dropTab;
 
-    struct DropTable {
-      Uint32 senderRef;
-      Uint32 senderData;
-      SignalCounter waitDropTabCount;
-    } m_prepDropTab;
-
     Uint8 kvalue;
     Uint8 noOfBackups;
-    Uint8 noPages;
+    Uint16 noPages;
     Uint16 tableType;
     Uint16 primaryTableId;
+
+    // set in local protocol during prepare until commit
+    Uint32 schemaTransId;
   };
   typedef Ptr<TabRecord> TabRecordPtr;
 
@@ -608,6 +679,7 @@ public:
     Uint32 fragId;
     TabRecordPtr rwfTabPtr;
     PageRecordPtr rwfPageptr;
+    Uint32 totalfragments;
   };
   struct CopyTableNode {
     Uint32 pageIndex;
@@ -618,12 +690,14 @@ public:
   };
   
 private:
+  friend class SimulatedBlock;
   BLOCK_DEFINES(Dbdih);
   
   void execDUMP_STATE_ORD(Signal *);
   void execNDB_TAMPER(Signal *);
   void execDEBUG_SIG(Signal *);
   void execEMPTY_LCP_CONF(Signal *);
+  void execEMPTY_LCP_REP(Signal*);
   void execMASTER_GCPREF(Signal *);
   void execMASTER_GCPREQ(Signal *);
   void execMASTER_GCPCONF(Signal *);
@@ -675,7 +749,9 @@ private:
   void execTCGETOPSIZECONF(Signal *);
   void execTC_CLOPSIZECONF(Signal *);
 
-  int handle_invalid_lcp_no(const class LcpFragRep*, ReplicaRecordPtr);
+  void execDIH_GET_TABINFO_REQ(Signal*);
+
+  int handle_invalid_lcp_no(const struct LcpFragRep*, ReplicaRecordPtr);
   void execLCP_FRAG_REP(Signal *);
   void execLCP_COMPLETE_REP(Signal *);
   void execSTART_LCP_REQ(Signal *);
@@ -723,13 +799,13 @@ private:
   void execPREPARE_COPY_FRAG_CONF(Signal*);
   void execDIADDTABREQ(Signal *);
   void execDIGETNODESREQ(Signal *);
-  void execDIRELEASEREQ(Signal *);
-  void execDISEIZEREQ(Signal *);
   void execSTTOR(Signal *);
-  void execDI_FCOUNTREQ(Signal *);
-  void execDIGETPRIMREQ(Signal *);
+  void execDIH_SCAN_TAB_REQ(Signal *);
+  void execDIH_SCAN_GET_NODES_REQ(Signal *);
+  void execDIH_SCAN_TAB_COMPLETE_REP(Signal*);
   void execGCP_SAVEREF(Signal *);
   void execGCP_TCFINISHED(Signal *);
+  void execGCP_TCFINISHED_sync_conf(Signal* signal, Uint32 cb, Uint32 err);
   void execREAD_NODESCONF(Signal *);
   void execNDB_STTOR(Signal *);
   void execDICTSTARTCONF(Signal *);
@@ -741,6 +817,8 @@ private:
   void execSTART_FRAGCONF(Signal *);
   void execADD_FRAGCONF(Signal *);
   void execADD_FRAGREF(Signal *);
+  void execDROP_FRAG_REF(Signal *);
+  void execDROP_FRAG_CONF(Signal *);
   void execFSOPENCONF(Signal *);
   void execFSOPENREF(Signal *);
   void execFSCLOSECONF(Signal *);
@@ -758,8 +836,6 @@ private:
   void execWAIT_GCP_CONF(Signal* signal);
 
   void execPREP_DROP_TAB_REQ(Signal* signal);
-  void execWAIT_DROP_TAB_REF(Signal* signal);
-  void execWAIT_DROP_TAB_CONF(Signal* signal);
   void execDROP_TAB_REQ(Signal* signal);
 
   void execALTER_TAB_REQ(Signal* signal);
@@ -767,14 +843,16 @@ private:
   void execCREATE_FRAGMENTATION_REQ(Signal*);
   
   void waitDropTabWritingToFile(Signal *, TabRecordPtr tabPtr);
-  void checkPrepDropTabComplete(Signal *, TabRecordPtr tabPtr);
-  void checkWaitDropTabFailedLqh(Signal *, Uint32 nodeId, Uint32 tableId);
+  void checkDropTabComplete(Signal *, TabRecordPtr tabPtr);
 
   void execDICT_LOCK_CONF(Signal* signal);
   void execDICT_LOCK_REF(Signal* signal);
 
   void execUPGRADE_PROTOCOL_ORD(Signal* signal);
 
+  void execCREATE_NODEGROUP_IMPL_REQ(Signal*);
+  void execDROP_NODEGROUP_IMPL_REQ(Signal*);
+
   // Statement blocks
 //------------------------------------
 // Methods that send signals
@@ -791,7 +869,7 @@ private:
   void sendINCL_NODEREQ(Signal *, Uint32 nodeId, Uint32);
   void sendMASTER_GCPREQ(Signal *, Uint32 nodeId, Uint32);
   void sendMASTER_LCPREQ(Signal *, Uint32 nodeId, Uint32);
-  void sendMASTER_LCPCONF(Signal * signal);
+  void sendMASTER_LCPCONF(Signal * signal, Uint32 fromLine);
   void sendSTART_RECREQ(Signal *, Uint32 nodeId, Uint32);
   void sendSTART_INFOREQ(Signal *, Uint32 nodeId, Uint32);
   void sendSTART_TOREQ(Signal *, Uint32 nodeId, Uint32);
@@ -820,7 +898,6 @@ private:
   void sendStartFragreq(Signal *,
                         TabRecordPtr regTabPtr,
                         Uint32 fragId);
-  void sendHOT_SPAREREP(Signal *);
   void sendAddFragreq(Signal *,
                       TabRecordPtr regTabPtr,
                       Uint32 fragId,
@@ -831,6 +908,7 @@ private:
   void addTable_closeConf(Signal* signal, Uint32 tabPtrI);
   void resetReplicaSr(TabRecordPtr tabPtr);
   void resetReplicaLcp(ReplicaRecord * replicaP, Uint32 stopGci);
+  void resetReplica(Ptr<ReplicaRecord>);
 
 //------------------------------------
 // Methods for LCP functionality
@@ -838,7 +916,7 @@ private:
   void checkKeepGci(TabRecordPtr, Uint32, Fragmentstore*, Uint32);
   void checkLcpStart(Signal *, Uint32 lineNo);
   void checkStartMoreLcp(Signal *, Uint32 nodeId);
-  bool reportLcpCompletion(const class LcpFragRep *);
+  bool reportLcpCompletion(const struct LcpFragRep *);
   void sendLCP_COMPLETE_REP(Signal *);
 
 //------------------------------------
@@ -890,7 +968,6 @@ private:
   void writeInitGcpErrorLab(Signal *, FileRecordPtr regFilePtr);
 
 
-  void calculateHotSpare();
   void checkEscalation();
   void clearRestartInfoBits(Signal *);
   void invalidateLcpInfoAfterSr(Signal*);
@@ -898,20 +975,22 @@ private:
   bool isMaster();
   bool isActiveMaster();
 
-  void emptyverificbuffer(Signal *, bool aContintueB);
-  Uint32 findHotSpare();
   void handleGcpStateInMaster(Signal *, NodeRecordPtr failedNodeptr);
   void initRestartInfo(Signal*);
   void initRestorableGciFiles();
   void makeNodeGroups(Uint32 nodeArray[]);
+  void add_nodegroup(NodeGroupRecordPtr);
+  void inc_ng_refcount(Uint32 ng);
+  void dec_ng_refcount(Uint32 ng);
+
   void makePrnList(class ReadNodesConf * readNodes, Uint32 nodeArray[]);
   void nodeResetStart(Signal* signal);
   void releaseTabPages(Uint32 tableId);
   void replication(Uint32 noOfReplicas,
                    NodeGroupRecordPtr NGPtr,
                    FragmentstorePtr regFragptr);
+  void sendDihRestartRef(Signal*);
   void selectMasterCandidateAndSend(Signal *);
-  void setInitialActiveStatus();
   void setLcpActiveStatusEnd(Signal*);
   void setLcpActiveStatusStart(Signal *);
   void setNodeActiveStatus();
@@ -950,6 +1029,8 @@ private:
   void checkLocalNodefailComplete(Signal*, Uint32 failedNodeId,
 				  NodefailHandlingStep step);
   
+  Callback m_sendSTTORRY;
+  void sendSTTORRY(Signal*, Uint32 senderData = 0, Uint32 retVal = 0);
   void ndbsttorry10Lab(Signal *, Uint32 _line);
   void createMutexes(Signal* signal, Uint32 no);
   void createMutex_done(Signal* signal, Uint32 no, Uint32 retVal);
@@ -1063,7 +1144,7 @@ private:
   void packCrashedReplicas(ReplicaRecordPtr pcrReplicaPtr);
   void releaseReplicas(Uint32 * replicaPtr);
   void removeOldCrashedReplicas(Uint32, Uint32, ReplicaRecordPtr rocReplicaPtr);
-  void removeTooNewCrashedReplicas(ReplicaRecordPtr rtnReplicaPtr);
+  void removeTooNewCrashedReplicas(ReplicaRecordPtr rtnReplicaPtr, Uint32 lastCompletedGCI);
   void mergeCrashedReplicas(ReplicaRecordPtr pcrReplicaPtr);
   void seizeReplicaRec(ReplicaRecordPtr& replicaPtr);
 
@@ -1074,7 +1155,9 @@ private:
   void allocStoredReplica(FragmentstorePtr regFragptr,
                           ReplicaRecordPtr& newReplicaPtr,
                           Uint32 nodeId);
-  Uint32 extractNodeInfo(const Fragmentstore * fragPtr, Uint32 nodes[]);
+  Uint32 extractNodeInfo(EmulatedJamBuffer *jambuf,
+                         const Fragmentstore * fragPtr,
+                         Uint32 nodes[]);
   bool findBestLogNode(CreateReplicaRecord* createReplica,
                        FragmentstorePtr regFragptr,
                        Uint32 startGci,
@@ -1117,6 +1200,18 @@ private:
   void getFragstore(TabRecord *, Uint32 fragNo, FragmentstorePtr & ptr);
   void initialiseFragstore();
 
+  void wait_old_scan(Signal*);
+  Uint32 add_fragments_to_table(Ptr<TabRecord>, const Uint16 buf[]);
+  Uint32 add_fragment_to_table(Ptr<TabRecord>, Uint32, Ptr<Fragmentstore>&);
+
+  void drop_fragments(Signal*, ConnectRecordPtr, Uint32 last);
+  void release_fragment_from_table(Ptr<TabRecord>, Uint32 fragId);
+  void send_alter_tab_ref(Signal*, Ptr<TabRecord>,Ptr<ConnectRecord>, Uint32);
+  void send_alter_tab_conf(Signal*, Ptr<ConnectRecord>);
+  void alter_table_writeTable_conf(Signal* signal, Uint32 ptrI, Uint32 err);
+  void saveTableFile(Signal*, Ptr<ConnectRecord>, Ptr<TabRecord>,
+                     TabRecord::CopyStatus, Callback&);
+
 //------------------------------------
 // Page Record specific methods
 //------------------------------------
@@ -1149,8 +1244,6 @@ private:
 //------------------------------------
   void changeNodeGroups(Uint32 startNode, Uint32 nodeTakenOver);
   void endTakeOver(Uint32 takeOverPtr);
-  void initStartTakeOver(const class StartToReq *, 
-			 TakeOverRecordPtr regTakeOverPtr);
   
   void systemRestartTakeOverLab(Signal *);
   void startTakeOver(Signal *,
@@ -1184,6 +1277,7 @@ private:
   void setAllowNodeStart(Uint32 nodeId, bool newState);
   bool getNodeCopyCompleted(Uint32 nodeId);
   void setNodeCopyCompleted(Uint32 nodeId, bool newState);
+  Uint32 getNodeGroup(Uint32 nodeId) const;
   bool checkNodeAlive(Uint32 nodeId);
 
   void nr_start_fragments(Signal*, TakeOverRecordPtr);
@@ -1191,13 +1285,18 @@ private:
   void nr_run_redo(Signal*, TakeOverRecordPtr);
   void nr_start_logging(Signal*, TakeOverRecordPtr);
 
+  void getTabInfo(Signal*);
+  void getTabInfo_send(Signal*, TabRecordPtr);
+  void getTabInfo_sendComplete(Signal*, Uint32, Uint32);
+  int getTabInfo_copyTableToSection(SegmentedSectionPtr & ptr, CopyTableNode);
+  int getTabInfo_copySectionToPages(TabRecordPtr, SegmentedSectionPtr);
+
   // Initialisation
   void initData();
   void initRecords();
 
   // Variables to support record structures and their free lists
 
-  ApiConnectRecord *apiConnectRecord;
   Uint32 capiConnectFileSize;
 
   ConnectRecord *connectRecord;
@@ -1214,11 +1313,12 @@ private:
   Fragmentstore *fragmentstore;
   Uint32 cfirstfragstore;
   Uint32 cfragstoreFileSize;
+  RSS_OP_SNAPSHOT(cremainingfrags);
 
   Uint32 c_nextNodeGroup;
   NodeGroupRecord *nodeGroupRecord;
-  Uint32 c_nextLogPart;
-  
+  RSS_OP_SNAPSHOT(cnghash);
+
   NodeRecord *nodeRecord;
 
   PageRecord *pageRecord;
@@ -1229,6 +1329,7 @@ private:
   Uint32 cfirstfreeReplica;
   Uint32 cnoFreeReplicaRec;
   Uint32 creplicaFileSize;
+  RSS_OP_SNAPSHOT(cnoFreeReplicaRec);
 
   TabRecord *tabRecord;
   Uint32 ctabFileSize;
@@ -1240,9 +1341,30 @@ private:
     2.4  C O M M O N    S T O R E D    V A R I A B L E S
     ----------------------------------------------------
   */
-  Uint32 cfirstVerifyQueue;
-  Uint32 clastVerifyQueue;
-  Uint32 cverifyQueueCounter;
+  struct DIVERIFY_queue
+  {
+    DIVERIFY_queue() {
+      m_ref = 0;
+      cfirstVerifyQueue = clastVerifyQueue = 0;
+      apiConnectRecord = 0;
+      m_empty_done = 1;
+    }
+    ApiConnectRecord *apiConnectRecord;
+    Uint32 cfirstVerifyQueue;
+    Uint32 clastVerifyQueue;
+    Uint32 m_empty_done;
+    Uint32 m_ref;
+    char pad[NDB_CL_PADSZ(sizeof(void*) + 4 * sizeof(Uint32))];
+  };
+
+  bool isEmpty(const DIVERIFY_queue&);
+  void enqueue(DIVERIFY_queue&, Uint32 senderData, Uint64 gci);
+  void dequeue(DIVERIFY_queue&, ApiConnectRecord &);
+  void emptyverificbuffer(Signal *, Uint32 q, bool aContintueB);
+  void emptyverificbuffer_check(Signal*, Uint32, Uint32);
+
+  DIVERIFY_queue c_diverify_queue[MAX_NDBMT_TC_THREADS];
+  Uint32 c_diverify_queue_cnt;
 
   /*------------------------------------------------------------------------*/
   /*       THIS VARIABLE KEEPS THE REFERENCES TO FILE RECORDS THAT DESCRIBE */
@@ -1284,8 +1406,15 @@ private:
    */
   struct MicroGcp
   {
+    MicroGcp() { }
     bool m_enabled;
     Uint32 m_master_ref;
+
+    /**
+     * rw-lock that protects multiple parallel DIVERIFY (readers) from
+     *   updates to gcp-state (e.g GCP_PREPARE, GCP_COMMIT)
+     */
+    NdbSeqLock m_lock;
     Uint64 m_old_gci;
     Uint64 m_current_gci; // Currently active
     Uint64 m_new_gci;     // Currently being prepared...
@@ -1485,6 +1614,11 @@ private:
   Uint32 c_set_initial_start_flag;
   Uint64 c_current_time; // Updated approx. every 10ms
 
+  /* Limit the number of concurrent table definition writes during LCP
+   * This avoids exhausting the DIH page pool
+   */
+  CountingSemaphore c_lcpTabDefWritesControl;
+
 public:
   enum LcpMasterTakeOverState {
     LMTOS_IDLE = 0,
@@ -1513,7 +1647,6 @@ private:
   } c_lcpMasterTakeOverState;
   
   Uint16 cmasterNodeId;
-  Uint8 cnoHotSpare;
 
   struct NodeStartMasterRecord {
     NodeStartMasterRecord() {}
@@ -1542,9 +1675,13 @@ private:
   Uint32 cnoReplicas;
 
   bool cwaitLcpSr;
+  /**
+   * Available nodegroups (ids) (length == cnoOfNodeGroups)
+   *   use to support nodegroups 2,4,6 (not just consequtive nodegroup ids)
+   */
+  Uint32 c_node_groups[MAX_NDB_NODES];
   Uint32 cnoOfNodeGroups;
   Uint32 crestartGci;      /* VALUE OF GCI WHEN SYSTEM RESTARTED OR STARTED */
-  Uint32 cminHotSpareNodes;
   
   /**
    * Counter variables keeping track of the number of outstanding signals
@@ -1657,11 +1794,13 @@ private:
    * Pool/list of WaitGCPMasterRecord record
    */
   ArrayPool<WaitGCPMasterRecord> waitGCPMasterPool;
-  DLList<WaitGCPMasterRecord> c_waitGCPMasterList;
+  typedef DLList<WaitGCPMasterRecord> WaitGCPList;
+  WaitGCPList c_waitGCPMasterList;
+  WaitGCPList c_waitEpochMasterList;
 
   void checkWaitGCPProxy(Signal*, NodeId failedNodeId);
   void checkWaitGCPMaster(Signal*, NodeId failedNodeId);
-  void emptyWaitGCPMasterQueue(Signal*);
+  void emptyWaitGCPMasterQueue(Signal*, Uint64, WaitGCPList&);
   
   /**
    * Stop me
@@ -1781,6 +1920,24 @@ private:
      */
     bool check_cut_log_tail(Uint32 gci) const;
   } m_local_lcp_state;
+
+  // MT LQH
+  Uint32 c_fragments_per_node_;
+  Uint32 getFragmentsPerNode();
+  Uint32 dihGetInstanceKey(FragmentstorePtr tFragPtr) {
+    ndbrequire(!tFragPtr.isNull());
+    Uint32 log_part_id = tFragPtr.p->m_log_part_id;
+    Uint32 instanceKey = 1 + (log_part_id % NDBMT_MAX_BLOCK_INSTANCES);
+    return instanceKey;
+  }
+  Uint32 dihGetInstanceKey(Uint32 tabId, Uint32 fragId);
+
+  /**
+   * Get minimum version of nodes in alive-list
+   */
+  Uint32 getMinVersion() const;
+
+  bool c_2pass_inr;
 };
 
 #if (DIH_CDATA_SIZE < _SYSFILE_SIZE32)
