@@ -32,33 +32,50 @@
 #include <signaldata/TupCommit.hpp>
 #include <signaldata/TupKey.hpp>
 #include <signaldata/NodeFailRep.hpp>
+#include <signaldata/NodeStateSignalData.hpp>
 
 #include <signaldata/DropTab.hpp>
 #include <SLList.hpp>
 
+#include <EventLogger.hpp>
+extern EventLogger * g_eventLogger;
+
 #define DEBUG(x) { ndbout << "TUP::" << x << endl; }
 
 void Dbtup::initData() 
 {
-  cnoOfAttrbufrec = ZNO_OF_ATTRBUFREC;
-  cnoOfFragrec = MAX_FRAG_PER_NODE;
-  cnoOfFragoprec = MAX_FRAG_PER_NODE;
-  cnoOfAlterTabOps = MAX_FRAG_PER_NODE;
+  TablerecPtr tablePtr;
+  cnoOfFragrec = NDB_ARRAY_SIZE(tablePtr.p->fragrec);
+  cnoOfFragoprec = NDB_ARRAY_SIZE(tablePtr.p->fragrec);
+  cnoOfAlterTabOps = NDB_ARRAY_SIZE(tablePtr.p->fragrec);
   c_maxTriggersPerTable = ZDEFAULT_MAX_NO_TRIGGERS_PER_TABLE;
   c_noOfBuildIndexRec = 32;
 
+  cCopyProcedure = RNIL;
+  cCopyLastSeg = RNIL;
+  cCopyOverwrite = 0;
+  cCopyOverwriteLen = 0;
+
   // Records with constant sizes
   init_list_sizes();
+  cpackedListIndex = 0;
+
+  m_minFreePages = 0;
 }//Dbtup::initData()
 
-Dbtup::Dbtup(Block_context& ctx, Pgman* pgman)
-  : SimulatedBlock(DBTUP, ctx),
+Dbtup::Dbtup(Block_context& ctx, Uint32 instanceNumber)
+  : SimulatedBlock(DBTUP, ctx, instanceNumber),
     c_lqh(0),
-    m_pgman(this, pgman),
+    c_tsman(0),
+    c_lgman(0),
+    c_pgman(0),
     c_extent_hash(c_extent_pool),
     c_storedProcPool(),
     c_buildIndexList(c_buildIndexPool),
-    c_undo_buffer(&ctx.m_mm)
+    c_undo_buffer(&ctx.m_mm),
+    m_pages_allocated(0),
+    m_pages_allocated_max(0),
+    f_undo_done(true)
 {
   BLOCK_CONSTRUCTOR(Dbtup);
 
@@ -68,14 +85,16 @@ Dbtup::Dbtup(Block_context& ctx, Pgman* pgman)
   addRecSignal(GSN_NODE_FAILREP, &Dbtup::execNODE_FAILREP);
 
   addRecSignal(GSN_DUMP_STATE_ORD, &Dbtup::execDUMP_STATE_ORD);
-  addRecSignal(GSN_SEND_PACKED, &Dbtup::execSEND_PACKED);
-  addRecSignal(GSN_ATTRINFO, &Dbtup::execATTRINFO);
+  addRecSignal(GSN_DBINFO_SCANREQ, &Dbtup::execDBINFO_SCANREQ);
+  addRecSignal(GSN_SEND_PACKED, &Dbtup::execSEND_PACKED, true);
   addRecSignal(GSN_STTOR, &Dbtup::execSTTOR);
   addRecSignal(GSN_MEMCHECKREQ, &Dbtup::execMEMCHECKREQ);
   addRecSignal(GSN_TUPKEYREQ, &Dbtup::execTUPKEYREQ);
   addRecSignal(GSN_TUPSEIZEREQ, &Dbtup::execTUPSEIZEREQ);
   addRecSignal(GSN_TUPRELEASEREQ, &Dbtup::execTUPRELEASEREQ);
   addRecSignal(GSN_STORED_PROCREQ, &Dbtup::execSTORED_PROCREQ);
+
+  addRecSignal(GSN_CREATE_TAB_REQ, &Dbtup::execCREATE_TAB_REQ);
   addRecSignal(GSN_TUPFRAGREQ, &Dbtup::execTUPFRAGREQ);
   addRecSignal(GSN_TUP_ADD_ATTRREQ, &Dbtup::execTUP_ADD_ATTRREQ);
   addRecSignal(GSN_ALTER_TAB_REQ, &Dbtup::execALTER_TAB_REQ);
@@ -83,10 +102,11 @@ Dbtup::Dbtup(Block_context& ctx, Pgman* pgman)
   addRecSignal(GSN_TUP_ABORTREQ, &Dbtup::execTUP_ABORTREQ);
   addRecSignal(GSN_NDB_STTOR, &Dbtup::execNDB_STTOR);
   addRecSignal(GSN_READ_CONFIG_REQ, &Dbtup::execREAD_CONFIG_REQ, true);
+  addRecSignal(GSN_NODE_STATE_REP, &Dbtup::execNODE_STATE_REP, true);
 
   // Trigger Signals
-  addRecSignal(GSN_CREATE_TRIG_REQ, &Dbtup::execCREATE_TRIG_REQ);
-  addRecSignal(GSN_DROP_TRIG_REQ,  &Dbtup::execDROP_TRIG_REQ);
+  addRecSignal(GSN_CREATE_TRIG_IMPL_REQ, &Dbtup::execCREATE_TRIG_IMPL_REQ);
+  addRecSignal(GSN_DROP_TRIG_IMPL_REQ,  &Dbtup::execDROP_TRIG_IMPL_REQ);
 
   addRecSignal(GSN_DROP_TAB_REQ, &Dbtup::execDROP_TAB_REQ);
 
@@ -94,9 +114,9 @@ Dbtup::Dbtup(Block_context& ctx, Pgman* pgman)
   addRecSignal(GSN_TUP_WRITELOG_REQ, &Dbtup::execTUP_WRITELOG_REQ);
 
   // Ordered index related
-  addRecSignal(GSN_BUILDINDXREQ, &Dbtup::execBUILDINDXREQ);
-  addRecSignal(GSN_BUILDINDXREF, &Dbtup::execBUILDINDXREF);
-  addRecSignal(GSN_BUILDINDXCONF, &Dbtup::execBUILDINDXCONF);
+  addRecSignal(GSN_BUILD_INDX_IMPL_REQ, &Dbtup::execBUILD_INDX_IMPL_REQ);
+  addRecSignal(GSN_BUILD_INDX_IMPL_REF, &Dbtup::execBUILD_INDX_IMPL_REF);
+  addRecSignal(GSN_BUILD_INDX_IMPL_CONF, &Dbtup::execBUILD_INDX_IMPL_CONF);
   addRecSignal(GSN_ALTER_TAB_CONF, &Dbtup::execALTER_TAB_CONF);
   m_max_parallel_index_build = 0;
 
@@ -112,17 +132,18 @@ Dbtup::Dbtup(Block_context& ctx, Pgman* pgman)
   addRecSignal(GSN_FSREMOVEREF, &Dbtup::execFSREMOVEREF, true);
   addRecSignal(GSN_FSREMOVECONF, &Dbtup::execFSREMOVECONF, true);
 
-  attrbufrec = 0;
+  addRecSignal(GSN_DROP_FRAG_REQ, &Dbtup::execDROP_FRAG_REQ);
+  addRecSignal(GSN_SUB_GCP_COMPLETE_REP, &Dbtup::execSUB_GCP_COMPLETE_REP);
+
+  addRecSignal(GSN_FIRE_TRIG_REQ, &Dbtup::execFIRE_TRIG_REQ);
+
   fragoperrec = 0;
   fragrecord = 0;
   alterTabOperRec = 0;
   hostBuffer = 0;
   tablerec = 0;
   tableDescriptor = 0;
-  totNoOfPagesAllocated = 0;
-  cnoOfAllocatedPages = 0;
-  c_no_of_pages = 0;
-  
+
   initData();
   CLEAR_ERROR_INSERT_VALUE;
 
@@ -130,15 +151,62 @@ Dbtup::Dbtup(Block_context& ctx, Pgman* pgman)
   RSS_OP_COUNTER_INIT(cnoOfFreeFragrec);
   RSS_OP_COUNTER_INIT(cnoOfFreeTabDescrRec);
   c_storedProcCountNonAPI = 0;
+
+  {
+    CallbackEntry& ce = m_callbackEntry[THE_NULL_CALLBACK];
+    ce.m_function = TheNULLCallback.m_callbackFunction;
+    ce.m_flags = 0;
+  }
+  { // 1
+    CallbackEntry& ce = m_callbackEntry[UNDO_CREATETABLE_LOGSYNC_CALLBACK];
+    ce.m_function = safe_cast(&Dbtup::undo_createtable_logsync_callback);
+    ce.m_flags = 0;
+  }
+  { // 2
+    CallbackEntry& ce = m_callbackEntry[DROP_TABLE_LOGSYNC_CALLBACK];
+    ce.m_function = safe_cast(&Dbtup::drop_table_logsync_callback);
+    ce.m_flags = 0;
+  }
+  { // 3
+    CallbackEntry& ce = m_callbackEntry[UNDO_CREATETABLE_CALLBACK];
+    ce.m_function = safe_cast(&Dbtup::undo_createtable_callback);
+    ce.m_flags = 0;
+  }
+  { // 4
+    CallbackEntry& ce = m_callbackEntry[DROP_TABLE_LOG_BUFFER_CALLBACK];
+    ce.m_function = safe_cast(&Dbtup::drop_table_log_buffer_callback);
+    ce.m_flags = 0;
+  }
+  { // 5
+    CallbackEntry& ce = m_callbackEntry[DROP_FRAGMENT_FREE_EXTENT_LOG_BUFFER_CALLBACK];
+    ce.m_function = safe_cast(&Dbtup::drop_fragment_free_extent_log_buffer_callback);
+    ce.m_flags = 0;
+  }
+  { // 6
+    CallbackEntry& ce = m_callbackEntry[NR_DELETE_LOG_BUFFER_CALLBACK];
+    ce.m_function = safe_cast(&Dbtup::nr_delete_log_buffer_callback);
+    ce.m_flags = 0;
+  }
+  { // 7
+    CallbackEntry& ce = m_callbackEntry[DISK_PAGE_LOG_BUFFER_CALLBACK];
+    ce.m_function = safe_cast(&Dbtup::disk_page_log_buffer_callback);
+    ce.m_flags = 0;
+  }
+  {
+    CallbackTable& ct = m_callbackTable;
+    ct.m_count = COUNT_CALLBACKS;
+    ct.m_entry = m_callbackEntry;
+    m_callbackTableAddr = &ct;
+  }
 }//Dbtup::Dbtup()
 
 Dbtup::~Dbtup() 
 {
+  /* Free Fragment Copy Procedure info */
+  freeCopyProcedure();
+
   // Records with dynamic sizes
   c_page_pool.clear();
-  deallocRecord((void **)&attrbufrec,"Attrbufrec", 
-		sizeof(Attrbufrec), 
-		cnoOfAttrbufrec);
   
   deallocRecord((void **)&fragoperrec,"Fragoperrec",
 		sizeof(Fragoperrec),
@@ -166,6 +234,19 @@ Dbtup::~Dbtup()
   
 }//Dbtup::~Dbtup()
 
+Dbtup::Apply_undo::Apply_undo()
+{
+  m_type = 0;
+  m_len = 0;
+  m_ptr = 0;
+  m_lsn = (Uint64)0;
+  m_table_ptr.setNull();
+  m_fragment_ptr.setNull();
+  m_page_ptr.setNull();
+  m_extent_ptr.setNull();
+  m_key.setNull();
+}
+
 BLOCK_FUNCTIONS(Dbtup)
 
 void Dbtup::execCONTINUEB(Signal* signal) 
@@ -173,6 +254,7 @@ void Dbtup::execCONTINUEB(Signal* signal)
   jamEntry();
   Uint32 actionType = signal->theData[0];
   Uint32 dataPtr = signal->theData[1];
+
   switch (actionType) {
   case ZINITIALISE_RECORDS:
     jam();
@@ -183,41 +265,6 @@ void Dbtup::execCONTINUEB(Signal* signal)
     jam();
     releaseFragment(signal, dataPtr, signal->theData[2]);
     break;
-  case ZREPORT_MEMORY_USAGE:{
-    jam();
-    static int c_currentMemUsed = 0;
-    Uint32 cnt = signal->theData[1];
-    Uint32 tmp = c_no_of_pages;
-    int now = tmp ? (cnoOfAllocatedPages * 100)/tmp : 0;
-    const int thresholds[] = { 100, 90, 80, 0 };
-    
-    Uint32 i = 0;
-    const Uint32 sz = sizeof(thresholds)/sizeof(thresholds[0]);
-    for(i = 0; i<sz; i++){
-      if(now >= thresholds[i]){
-	now = thresholds[i];
-	break;
-      }
-    }
-
-    if(now != c_currentMemUsed || 
-       (c_memusage_report_frequency && cnt + 1 == c_memusage_report_frequency))
-    {
-      reportMemoryUsage(signal, 
-			now > c_currentMemUsed ? 1 : 
-			now < c_currentMemUsed ? -1 : 0);
-      cnt = 0;
-      c_currentMemUsed = now;
-    } 
-    else
-    {
-      cnt++;
-    }
-    signal->theData[0] = ZREPORT_MEMORY_USAGE;
-    signal->theData[1] = cnt;
-    sendSignalWithDelay(reference(), GSN_CONTINUEB, signal, 1000, 2);    
-    return;
-  }
   case ZBUILD_INDEX:
     jam();
     buildIndex(signal, dataPtr);
@@ -233,7 +280,6 @@ void Dbtup::execCONTINUEB(Signal* signal)
   case ZFREE_EXTENT:
   {
     jam();
-    
     TablerecPtr tabPtr;
     tabPtr.i= dataPtr;
     FragrecordPtr fragPtr;
@@ -246,7 +292,6 @@ void Dbtup::execCONTINUEB(Signal* signal)
   case ZUNMAP_PAGES:
   {
     jam();
-    
     TablerecPtr tabPtr;
     tabPtr.i= dataPtr;
     FragrecordPtr fragPtr;
@@ -274,6 +319,27 @@ void Dbtup::execCONTINUEB(Signal* signal)
     rebuild_page_free_list(signal);
     return;
   }
+  case ZDISK_RESTART_UNDO:
+  {
+    jam();
+    if (!assembleFragments(signal)) {
+      jam();
+      return;
+    }
+    Uint32 type = signal->theData[1];
+    Uint32 len = signal->theData[2];
+    Uint64 lsn_hi = signal->theData[3];
+    Uint64 lsn_lo = signal->theData[4];
+    Uint64 lsn = (lsn_hi << 32) | lsn_lo;
+    SectionHandle handle(this, signal);
+    ndbrequire(handle.m_cnt == 1);
+    SegmentedSectionPtr ssptr;
+    handle.getSection(ssptr, 0);
+    ::copy(c_proxy_undo_data, ssptr);
+    releaseSections(handle);
+    disk_restart_undo(signal, lsn, type, c_proxy_undo_data, len);
+    return;
+  }
 
   default:
     ndbrequire(false);
@@ -294,10 +360,11 @@ void Dbtup::execSTTOR(Signal* signal)
   switch (startPhase) {
   case ZSTARTPHASE1:
     jam();
-    ndbrequire((c_lqh= (Dblqh*)globalData.getBlock(DBLQH)) != 0);
+    ndbrequire((c_lqh= (Dblqh*)globalData.getBlock(DBLQH, instance())) != 0);
     ndbrequire((c_tsman= (Tsman*)globalData.getBlock(TSMAN)) != 0);
     ndbrequire((c_lgman= (Lgman*)globalData.getBlock(LGMAN)) != 0);
-    cownref = calcTupBlockRef(0);
+    ndbrequire((c_pgman= (Pgman*)globalData.getBlock(PGMAN, instance())) != 0);
+    cownref = calcInstanceBlockRef(DBTUP);
     break;
   default:
     jam();
@@ -308,7 +375,8 @@ void Dbtup::execSTTOR(Signal* signal)
   signal->theData[2] = 2;
   signal->theData[3] = ZSTARTPHASE1;
   signal->theData[4] = 255;
-  sendSignal(NDBCNTR_REF, GSN_STTORRY, signal, 5, JBB);
+  BlockReference cntrRef = !isNdbMtLqh() ? NDBCNTR_REF : DBTUP_REF;
+  sendSignal(cntrRef, GSN_STTORRY, signal, 5, JBB);
   return;
 }//Dbtup::execSTTOR()
 
@@ -332,21 +400,51 @@ void Dbtup::execREAD_CONFIG_REQ(Signal* signal)
   ndbrequire(!ndb_mgm_get_int_parameter(p, CFG_TUP_FRAG, &cnoOfFragrec));
   
   Uint32 noOfTriggers= 0;
+  Uint32 noOfAttribs = 0;
   
   ndbrequire(!ndb_mgm_get_int_parameter(p, CFG_TUP_TABLE, &cnoOfTablerec));
-  ndbrequire(!ndb_mgm_get_int_parameter(p, CFG_TUP_TABLE_DESC, 
-					&cnoOfTabDescrRec));
+  ndbrequire(!ndb_mgm_get_int_parameter(p, CFG_DB_NO_ATTRIBUTES, &noOfAttribs));
+
   Uint32 noOfStoredProc;
   ndbrequire(!ndb_mgm_get_int_parameter(p, CFG_TUP_STORED_PROC, 
 					&noOfStoredProc));
   ndbrequire(!ndb_mgm_get_int_parameter(p, CFG_DB_NO_TRIGGERS, 
 					&noOfTriggers));
 
-  cnoOfTabDescrRec = (cnoOfTabDescrRec & 0xFFFFFFF0) + 16;
+
+  {
+    Uint32 keyDesc = noOfAttribs;
+    Uint32 maxKeyDesc = cnoOfTablerec * MAX_ATTRIBUTES_IN_INDEX;
+    if (keyDesc > maxKeyDesc)
+    {
+      /**
+       * There can be no-more key's
+       *   than "cnoOfTablerec * MAX_ATTRIBUTES_IN_INDEX"
+       */
+      jam();
+      keyDesc = maxKeyDesc;
+    }
+
+    cnoOfTabDescrRec =
+      cnoOfTablerec * 2 * (ZTD_SIZE + ZTD_TRAILER_SIZE) +
+      noOfAttribs * (sizeOfReadFunction() + // READ
+                     sizeOfReadFunction() + // UPDATE
+                     (sizeof(char*) >> 2) + // Charset
+                     ZAD_SIZE +             // Descriptor
+                     1 +                    // real order
+                     InternalMaxDynFix) +   // Worst case dynamic
+      keyDesc;                              // key-descr
+
+    cnoOfTabDescrRec = (cnoOfTabDescrRec & 0xFFFFFFF0) + 16;
+  }
 
   initRecords();
 
   c_storedProcPool.setSize(noOfStoredProc);
+
+  // Allocate fragment copy procedure
+  allocCopyProcedure();
+
   c_buildIndexPool.setSize(c_noOfBuildIndexRec);
   c_triggerPool.setSize(noOfTriggers, false, true, true, CFG_DB_NO_TRIGGERS);
 
@@ -371,10 +469,12 @@ void Dbtup::execREAD_CONFIG_REQ(Signal* signal)
   {
     Uint64 tmp = 64*1024*1024;
     ndb_mgm_get_int64_parameter(p, CFG_DB_DISK_PAGE_BUFFER_MEMORY, &tmp);
-    m_max_page_read_ahead = (tmp  + GLOBAL_PAGE_SIZE - 1) / GLOBAL_PAGE_SIZE; // in pages
+    tmp = (tmp  + GLOBAL_PAGE_SIZE - 1) / GLOBAL_PAGE_SIZE; // in pages
     // never read ahead more than 32 pages
-    if (m_max_page_read_ahead > 32)
+    if (tmp > 32)
       m_max_page_read_ahead = 32;
+    else
+      m_max_page_read_ahead = (Uint32)tmp;
   }
 
 
@@ -388,14 +488,28 @@ void Dbtup::execREAD_CONFIG_REQ(Signal* signal)
   clastBitMask = 1;
   clastBitMask = clastBitMask << 31;
 
-  c_memusage_report_frequency = 0;
-  ndb_mgm_get_int_parameter(p, CFG_DB_MEMREPORT_FREQUENCY, 
-			    &c_memusage_report_frequency);
-
   ndb_mgm_get_int_parameter(p, CFG_DB_MT_BUILD_INDEX,
                             &m_max_parallel_index_build);
+
+  if (isNdbMtLqh() && globalData.ndbMtLqhThreads > 1)
+  {
+    /**
+     * Divide by LQH threads
+     */
+    Uint32 val = m_max_parallel_index_build;
+    val = (val + instance() - 1) / globalData.ndbMtLqhThreads;
+    m_max_parallel_index_build = val;
+  }
   
   initialiseRecordsLab(signal, 0, ref, senderData);
+
+  {
+    Uint32 val = 0;
+    ndb_mgm_get_int_parameter(p, CFG_DB_CRASH_ON_CORRUPTED_TUPLE,
+                              &val);
+    c_crashOnCorruptedTuple = val ? true : false;
+  }
+
 }//Dbtup::execSIZEALT_REP()
 
 void Dbtup::initRecords() 
@@ -407,16 +521,9 @@ void Dbtup::initRecords()
     m_ctx.m_config.getOwnConfigIterator();
   ndbrequire(p != 0);
 
-  ndbrequire(!ndb_mgm_get_int_parameter(p, CFG_TUP_PAGE, &tmp));
-
   // Records with dynamic sizes
   void* ptr = m_ctx.m_mm.get_memroot();
   c_page_pool.set((Page*)ptr, (Uint32)~0);
-  c_no_of_pages = tmp;
-
-  attrbufrec = (Attrbufrec*)allocRecord("Attrbufrec", 
-					sizeof(Attrbufrec), 
-					cnoOfAttrbufrec);
 
   fragoperrec = (Fragoperrec*)allocRecord("Fragoperrec",
 					  sizeof(Fragoperrec),
@@ -503,7 +610,6 @@ void Dbtup::initialiseRecordsLab(Signal* signal, Uint32 switchData,
     break;
   case 12:
     jam();
-    initializeAttrbufrec();
     break;
   case 13:
     jam();
@@ -541,7 +647,8 @@ void Dbtup::execNDB_STTOR(Signal* signal)
   case ZSTARTPHASE1:
     jam();
     cownNodeId = ownNodeId;
-    cownref = calcTupBlockRef(ownNodeId);
+    cownref = calcInstanceBlockRef(DBTUP);
+    initializeDefaultValuesFrag();
     break;
   case ZSTARTPHASE2:
     jam();
@@ -555,42 +662,42 @@ void Dbtup::execNDB_STTOR(Signal* signal)
     break;
   case ZSTARTPHASE6:
     jam();
-/*****************************************/
-/*       NOW SET THE DISK WRITE SPEED TO */
-/*       PAGES PER TICK AFTER SYSTEM     */
-/*       RESTART.                        */
-/*****************************************/
-    signal->theData[0] = ZREPORT_MEMORY_USAGE;
-    signal->theData[1] = 0;
-    sendSignalWithDelay(reference(), GSN_CONTINUEB, signal, 1000, 2);
     break;
   default:
     jam();
     break;
   }//switch
   signal->theData[0] = cownref;
-  sendSignal(cndbcntrRef, GSN_NDB_STTORRY, signal, 1, JBB);
+  BlockReference cntrRef = !isNdbMtLqh() ? NDBCNTR_REF : DBTUP_REF;
+  sendSignal(cntrRef, GSN_NDB_STTORRY, signal, 1, JBB);
 }//Dbtup::execNDB_STTOR()
 
 void Dbtup::startphase3Lab(Signal* signal, Uint32 config1, Uint32 config2) 
 {
 }//Dbtup::startphase3Lab()
 
-void Dbtup::initializeAttrbufrec() 
+void Dbtup::initializeDefaultValuesFrag()
 {
-  AttrbufrecPtr attrBufPtr;
-  for (attrBufPtr.i = 0;
-       attrBufPtr.i < cnoOfAttrbufrec; attrBufPtr.i++) {
-    refresh_watch_dog();
-    ptrAss(attrBufPtr, attrbufrec);
-    attrBufPtr.p->attrbuf[ZBUF_NEXT] = attrBufPtr.i + 1;
-  }//for
-  attrBufPtr.i = cnoOfAttrbufrec - 1;
-  ptrAss(attrBufPtr, attrbufrec);
-  attrBufPtr.p->attrbuf[ZBUF_NEXT] = RNIL;
-  cfirstfreeAttrbufrec = 0;
-  cnoFreeAttrbufrec = cnoOfAttrbufrec;
-}//Dbtup::initializeAttrbufrec()
+  /* Grab and initialize a fragment record for storing default
+   * values for the table fragments held by this TUP instance
+   */
+  seizeFragrecord(DefaultValuesFragment);
+  DefaultValuesFragment.p->fragStatus = Fragrecord::FS_ONLINE;
+  DefaultValuesFragment.p->m_undo_complete= false;
+  DefaultValuesFragment.p->m_lcp_scan_op = RNIL;
+  DefaultValuesFragment.p->noOfPages = 0;
+  DefaultValuesFragment.p->noOfVarPages = 0;
+  DefaultValuesFragment.p->m_max_page_no = 0;
+  DefaultValuesFragment.p->m_free_page_id_list = FREE_PAGE_RNIL;
+  ndbrequire(DefaultValuesFragment.p->m_page_map.isEmpty());
+  DefaultValuesFragment.p->m_restore_lcp_id = RNIL;
+  for (Uint32 i = 0; i<MAX_FREE_LIST+1; i++)
+    ndbrequire(DefaultValuesFragment.p->free_var_page_array[i].isEmpty());
+
+  DefaultValuesFragment.p->m_logfile_group_id = RNIL;
+
+  return;
+}
 
 void Dbtup::initializeFragoperrec() 
 {
@@ -613,7 +720,7 @@ void Dbtup::initializeFragrecord()
     ptrAss(regFragPtr, fragrecord);
     new (regFragPtr.p) Fragrecord();
     regFragPtr.p->nextfreefrag = regFragPtr.i + 1;
-    regFragPtr.p->fragStatus = IDLE;
+    regFragPtr.p->fragStatus = Fragrecord::FS_FREE;
   }//for
   regFragPtr.i = cnoOfFragrec - 1;
   ptrAss(regFragPtr, fragrecord);
@@ -670,7 +777,7 @@ void Dbtup::initializeTablerec()
 void
 Dbtup::initTab(Tablerec* const regTabPtr)
 {
-  for (Uint32 i = 0; i < MAX_FRAG_PER_NODE; i++) {
+  for (Uint32 i = 0; i < NDB_ARRAY_SIZE(regTabPtr->fragid); i++) {
     regTabPtr->fragid[i] = RNIL;
     regTabPtr->fragrec[i] = RNIL;
   }//for
@@ -680,7 +787,12 @@ Dbtup::initTab(Tablerec* const regTabPtr)
 
   regTabPtr->tabDescriptor = RNIL;
   regTabPtr->readKeyArray = RNIL;
-  regTabPtr->dynTabDescriptor = RNIL;
+  regTabPtr->dynTabDescriptor[MM] = RNIL;
+  regTabPtr->dynTabDescriptor[DD] = RNIL;
+  regTabPtr->dynFixSizeMask[MM] = NULL;
+  regTabPtr->dynVarSizeMask[MM] = NULL;
+  regTabPtr->dynFixSizeMask[DD] = NULL;
+  regTabPtr->dynVarSizeMask[DD] = NULL;
 
   regTabPtr->m_bits = 0;
 
@@ -690,6 +802,7 @@ Dbtup::initTab(Tablerec* const regTabPtr)
   regTabPtr->m_dropTable.tabUserPtr = RNIL;
   regTabPtr->m_dropTable.tabUserRef = 0;
   regTabPtr->tableStatus = NOT_DEFINED;
+  regTabPtr->m_default_value_location.setNull();
 
   // Clear trigger data
   if (!regTabPtr->afterInsertTriggers.isEmpty())
@@ -745,13 +858,10 @@ void Dbtup::execTUPSEIZEREQ(Signal* signal)
   }//if
 
   new (regOperPtr.p) Operationrec();
-  regOperPtr.p->firstAttrinbufrec = RNIL;
-  regOperPtr.p->lastAttrinbufrec = RNIL;
   regOperPtr.p->m_any_value = 0;
   regOperPtr.p->op_struct.op_type = ZREAD;
   regOperPtr.p->op_struct.in_active_list = false;
   set_trans_state(regOperPtr.p, TRANS_DISCONNECTED);
-  regOperPtr.p->storedProcedureId = ZNIL;
   regOperPtr.p->prevActiveOp = RNIL;
   regOperPtr.p->nextActiveOp = RNIL;
   regOperPtr.p->tupVersion = ZNIL;
@@ -765,7 +875,7 @@ void Dbtup::execTUPSEIZEREQ(Signal* signal)
   return;
 }//Dbtup::execTUPSEIZEREQ()
 
-#define printFragment(t){ for(Uint32 i = 0; i < MAX_FRAG_PER_NODE;i++){\
+#define printFragment(t){ for(Uint32 i = 0; i < NDB_ARRAY_SIZE(t.p->fragid);i++){ \
   ndbout_c("table = %d fragid[%d] = %d fragrec[%d] = %d", \
            t.i, t.p->fragid[i], i, t.p->fragrec[i]); }}
 
@@ -810,3 +920,31 @@ void Dbtup::execNODE_FAILREP(Signal* signal)
   }//for
 }
 
+extern Uint32 compute_acc_32kpages(const ndb_mgm_configuration_iterator * p);
+
+void
+Dbtup::execNODE_STATE_REP(Signal* signal)
+{
+  jamEntry();
+  const NodeStateRep* rep = CAST_CONSTPTR(NodeStateRep,
+                                          signal->getDataPtr());
+
+  if (rep->nodeState.startLevel == NodeState::SL_STARTED)
+  {
+    jam();
+
+    const ndb_mgm_configuration_iterator * p =
+      m_ctx.m_config.getOwnConfigIterator();
+    ndbrequire(p != 0);
+
+    Uint32 free_pct = 5;
+    ndb_mgm_get_int_parameter(p, CFG_DB_FREE_PCT, &free_pct);
+
+    Uint32 accpages = compute_acc_32kpages(p);
+
+    Resource_limit rl;
+    m_ctx.m_mm.get_resource_limit(RG_DATAMEM, rl);
+    m_minFreePages = ((rl.m_min - accpages) * free_pct) / 100;
+  }
+  SimulatedBlock::execNODE_STATE_REP(signal);
+}
