@@ -1,5 +1,5 @@
 /*
-   Copyright (c) 2003, 2010, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2003, 2011, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -30,26 +30,25 @@ Dbtux::treeAdd(TuxCtx& ctx, Frag& frag, TreePos treePos, TreeEnt ent)
   do {
     if (treePos.m_loc != NullTupLoc) {
       // non-empty tree
-      thrjam(ctx.jambase, ctx.jamidx);
+      thrjam(ctx.jamBuffer);
       selectNode(node, treePos.m_loc);
       unsigned pos = treePos.m_pos;
       if (node.getOccup() < tree.m_maxOccup) {
         // node has room
-        thrjam(ctx.jambase, ctx.jamidx);
+        thrjam(ctx.jamBuffer);
         nodePushUp(ctx, node, pos, ent, RNIL);
         break;
       }
       treeAddFull(ctx, frag, node, pos, ent);
       break;
     }
-    thrjam(ctx.jambase, ctx.jamidx);
+    thrjam(ctx.jamBuffer);
     insertNode(node);
     nodePushUp(ctx, node, 0, ent, RNIL);
     node.setSide(2);
     tree.m_root = node.m_loc;
     break;
   } while (0);
-  tree.m_entryCount++;
 }
 
 /*
@@ -66,16 +65,16 @@ Dbtux::treeAddFull(TuxCtx& ctx, Frag& frag, NodeHandle lubNode, unsigned pos, Tr
     // find g.l.b node
     NodeHandle glbNode(frag);
     do {
-      thrjam(ctx.jambase, ctx.jamidx);
+      thrjam(ctx.jamBuffer);
       selectNode(glbNode, loc);
       loc = glbNode.getLink(1);
     } while (loc != NullTupLoc);
     if (glbNode.getOccup() < tree.m_maxOccup) {
       // g.l.b node has room
-      thrjam(ctx.jambase, ctx.jamidx);
+      thrjam(ctx.jamBuffer);
       Uint32 scanList = RNIL;
       if (pos != 0) {
-        thrjam(ctx.jambase, ctx.jamidx);
+        thrjam(ctx.jamBuffer);
         // add the new entry and return min entry
         nodePushDown(ctx, lubNode, pos - 1, ent, scanList);
       }
@@ -106,7 +105,7 @@ Dbtux::treeAddNode(TuxCtx& ctx,
   glbNode.setSide(i);
   Uint32 scanList = RNIL;
   if (pos != 0) {
-    thrjam(ctx.jambase, ctx.jamidx);
+    thrjam(ctx.jamBuffer);
     // add the new entry and return min entry
     nodePushDown(ctx, lubNode, pos - 1, ent, scanList);
   }
@@ -129,26 +128,26 @@ Dbtux::treeAddRebalance(TuxCtx & ctx, Frag& frag, NodeHandle node, unsigned i)
     int b = node.getBalance();
     if (b == 0) {
       // perfectly balanced
-      thrjam(ctx.jambase, ctx.jamidx);
+      thrjam(ctx.jamBuffer);
       node.setBalance(j);
       // height change propagates up
     } else if (b == -j) {
       // height of shorter subtree increased
-      thrjam(ctx.jambase, ctx.jamidx);
+      thrjam(ctx.jamBuffer);
       node.setBalance(0);
       // height of tree did not change - done
       break;
     } else if (b == j) {
       // height of longer subtree increased
-      thrjam(ctx.jambase, ctx.jamidx);
+      thrjam(ctx.jamBuffer);
       NodeHandle childNode(frag);
       selectNode(childNode, node.getLink(i));
       int b2 = childNode.getBalance();
       if (b2 == b) {
-        thrjam(ctx.jambase, ctx.jamidx);
+        thrjam(ctx.jamBuffer);
         treeRotateSingle(ctx, frag, node, i);
       } else if (b2 == -b) {
-        thrjam(ctx.jambase, ctx.jamidx);
+        thrjam(ctx.jamBuffer);
         treeRotateDouble(ctx, frag, node, i);
       } else {
         // height of subtree increased so it cannot be perfectly balanced
@@ -161,7 +160,7 @@ Dbtux::treeAddRebalance(TuxCtx & ctx, Frag& frag, NodeHandle node, unsigned i)
     }
     TupLoc parentLoc = node.getLink(2);
     if (parentLoc == NullTupLoc) {
-      thrjam(ctx.jambase, ctx.jamidx);
+      thrjam(ctx.jamBuffer);
       // root node - done
       break;
     }
@@ -212,8 +211,6 @@ Dbtux::treeRemove(Frag& frag, TreePos treePos)
     treeRemoveLeaf(frag, node);
     break;
   } while (0);
-  ndbrequire(tree.m_entryCount != 0);
-  tree.m_entryCount--;
 }
 
 /*
@@ -441,7 +438,7 @@ Dbtux::treeRotateSingle(TuxCtx& ctx, Frag& frag, NodeHandle& node, unsigned i)
   TupLoc loc4 = node3.getLink(1 - i);
   NodeHandle node4(frag);
   if (loc4 != NullTupLoc) {
-    thrjam(ctx.jambase, ctx.jamidx);
+    thrjam(ctx.jamBuffer);
     selectNode(node4, loc4);
     ndbrequire(node4.getSide() == (1 - i) &&
                node4.getLink(2) == loc3);
@@ -475,12 +472,12 @@ Dbtux::treeRotateSingle(TuxCtx& ctx, Frag& frag, NodeHandle& node, unsigned i)
   node5.setLink(2, loc3);
   node5.setSide(1 - i);
   if (loc0 != NullTupLoc) {
-    thrjam(ctx.jambase, ctx.jamidx);
+    thrjam(ctx.jamBuffer);
     NodeHandle node0(frag);
     selectNode(node0, loc0);
     node0.setLink(side5, loc3);
   } else {
-    thrjam(ctx.jambase, ctx.jamidx);
+    thrjam(ctx.jamBuffer);
     frag.m_tree.m_root = loc3;
   }//if
   /* The final step of the change is to update the balance of 3 and
@@ -495,11 +492,11 @@ Dbtux::treeRotateSingle(TuxCtx& ctx, Frag& frag, NodeHandle& node, unsigned i)
   be unbalanced in the opposite direction of 5.
   */
   if (bal3 == bal5) {
-    thrjam(ctx.jambase, ctx.jamidx);
+    thrjam(ctx.jamBuffer);
     node3.setBalance(0);
     node5.setBalance(0);
   } else if (bal3 == 0) {
-    thrjam(ctx.jambase, ctx.jamidx);
+    thrjam(ctx.jamBuffer);
     node3.setBalance(-bal5);
     node5.setBalance(bal5);
   } else {
@@ -650,9 +647,9 @@ Dbtux::treeRotateDouble(TuxCtx& ctx, Frag& frag, NodeHandle& node, unsigned i)
 
   // fill up leaf before it becomes internal
   if (loc3 == NullTupLoc && loc5 == NullTupLoc) {
-    thrjam(ctx.jambase, ctx.jamidx);
+    thrjam(ctx.jamBuffer);
     if (node4.getOccup() < tree.m_minOccup) {
-      thrjam(ctx.jambase, ctx.jamidx);
+      thrjam(ctx.jamBuffer);
       unsigned cnt = tree.m_minOccup - node4.getOccup();
       ndbrequire(cnt < node2.getOccup());
       nodeSlide(ctx, node4, node2, cnt, i);
@@ -661,14 +658,14 @@ Dbtux::treeRotateDouble(TuxCtx& ctx, Frag& frag, NodeHandle& node, unsigned i)
     }
   } else {
     if (loc3 != NullTupLoc) {
-      thrjam(ctx.jambase, ctx.jamidx);
+      thrjam(ctx.jamBuffer);
       NodeHandle node3(frag);
       selectNode(node3, loc3);
       node3.setLink(2, loc2);
       node3.setSide(1 - i);
     }
     if (loc5 != NullTupLoc) {
-      thrjam(ctx.jambase, ctx.jamidx);
+      thrjam(ctx.jamBuffer);
       NodeHandle node5(frag);
       selectNode(node5, loc5);
       node5.setLink(2, node6.m_loc);
@@ -692,25 +689,25 @@ Dbtux::treeRotateDouble(TuxCtx& ctx, Frag& frag, NodeHandle& node, unsigned i)
   node4.setSide(side6);
 
   if (loc0 != NullTupLoc) {
-    thrjam(ctx.jambase, ctx.jamidx);
+    thrjam(ctx.jamBuffer);
     selectNode(node0, loc0);
     node0.setLink(side6, loc4);
   } else {
-    thrjam(ctx.jambase, ctx.jamidx);
+    thrjam(ctx.jamBuffer);
     frag.m_tree.m_root = loc4;
   }
   // set balance of changed nodes
   node4.setBalance(0);
   if (bal4 == 0) {
-    thrjam(ctx.jambase, ctx.jamidx);
+    thrjam(ctx.jamBuffer);
     node2.setBalance(0);
     node6.setBalance(0);
   } else if (bal4 == -bal2) {
-    thrjam(ctx.jambase, ctx.jamidx);
+    thrjam(ctx.jamBuffer);
     node2.setBalance(0);
     node6.setBalance(bal2);
   } else if (bal4 == bal2) {
-    thrjam(ctx.jambase, ctx.jamidx);
+    thrjam(ctx.jamBuffer);
     node2.setBalance(-bal2);
     node6.setBalance(0);
   } else {
