@@ -1,4 +1,5 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2003, 2011, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,29 +12,32 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 #include <ndb_global.h>
+#include <ndb_version.h>
 
 #include "InitConfigFileParser.hpp"
 #include "Config.hpp"
-#include "MgmtErrorReporter.hpp"
 #include <NdbOut.hpp>
 #include "ConfigInfo.hpp"
+#include "EventLogger.hpp"
 #include <m_string.h>
+#include <util/SparseBitmask.hpp>
+#include "../common/util/parse_mask.hpp"
+
+extern EventLogger *g_eventLogger;
 
 const int MAX_LINE_LENGTH = 1024;  // Max length of line of text in config file
 static void trim(char *);
 
-static void require(bool v) { if(!v) abort();}
-
 //****************************************************************************
 //  Ctor / Dtor
 //****************************************************************************
-InitConfigFileParser::InitConfigFileParser(FILE * out)
+InitConfigFileParser::InitConfigFileParser()
 {
   m_info = new ConfigInfo();
-  m_errstream = out ? out : stdout;
 }
 
 InitConfigFileParser::~InitConfigFileParser() {
@@ -43,12 +47,11 @@ InitConfigFileParser::~InitConfigFileParser() {
 //****************************************************************************
 //  Read Config File
 //****************************************************************************
-InitConfigFileParser::Context::Context(const ConfigInfo * info, FILE * out)
+InitConfigFileParser::Context::Context(const ConfigInfo * info)
   :  m_userProperties(true), m_configValues(1000, 20) {
 
   m_config = new Properties(true);
   m_defaults = new Properties(true);
-  m_errstream = out;
 }
 
 InitConfigFileParser::Context::~Context(){
@@ -63,7 +66,7 @@ Config *
 InitConfigFileParser::parseConfig(const char * filename) {
   FILE * file = fopen(filename, "r");
   if(file == 0){
-    fprintf(m_errstream, "Error opening file: %s\n", filename);
+    g_eventLogger->error("Error opening '%s', error: %d, %s", filename, errno, strerror(errno));
     return 0;
   }
   
@@ -77,7 +80,7 @@ InitConfigFileParser::parseConfig(FILE * file) {
 
   char line[MAX_LINE_LENGTH];
 
-  Context ctx(m_info, m_errstream); 
+  Context ctx(m_info);
   ctx.m_lineno = 0;
   ctx.m_currentSection = 0;
 
@@ -113,7 +116,8 @@ InitConfigFileParser::parseConfig(FILE * file) {
 			"of configuration file.");
 	return 0;
       }
-      BaseString::snprintf(ctx.fname, sizeof(ctx.fname), section); free(section);
+      BaseString::snprintf(ctx.fname, sizeof(ctx.fname), "%s", section);
+      free(section);
       ctx.type             = InitConfigFileParser::DefaultSection;
       ctx.m_sectionLineno  = ctx.m_lineno;
       ctx.m_currentSection = new Properties(true);
@@ -133,7 +137,7 @@ InitConfigFileParser::parseConfig(FILE * file) {
 			"of configuration file.");
 	return 0;
       }
-      BaseString::snprintf(ctx.fname, sizeof(ctx.fname), section);
+      BaseString::snprintf(ctx.fname, sizeof(ctx.fname), "%s", section);
       free(section);
       ctx.type             = InitConfigFileParser::Section;
       ctx.m_sectionLineno  = ctx.m_lineno;      
@@ -171,6 +175,7 @@ InitConfigFileParser::run_config_rules(Context& ctx)
 {
   for(size_t i = 0; ConfigInfo::m_ConfigRules[i].m_configRule != 0; i++){
     ctx.type             = InitConfigFileParser::Undefined;
+    ctx.m_info           = m_info;
     ctx.m_currentSection = 0;
     ctx.m_userDefaults   = 0;
     ctx.m_currentInfo    = 0;
@@ -181,8 +186,9 @@ InitConfigFileParser::run_config_rules(Context& ctx)
 						      ConfigInfo::m_ConfigRules[i].m_ruleData))
       return 0;
 
-    for(size_t j = 0; j<tmp.size(); j++){
-      BaseString::snprintf(ctx.fname, sizeof(ctx.fname), tmp[j].m_sectionType.c_str());
+    for(unsigned j = 0; j<tmp.size(); j++){
+      BaseString::snprintf(ctx.fname, sizeof(ctx.fname),
+                           "%s", tmp[j].m_sectionType.c_str());
       ctx.type             = InitConfigFileParser::Section;
       ctx.m_currentSection = tmp[j].m_sectionData;
       ctx.m_userDefaults   = getSection(ctx.fname, ctx.m_defaults);
@@ -208,15 +214,11 @@ InitConfigFileParser::run_config_rules(Context& ctx)
   ctx.m_config->put("NoOfNodes", nNodes);
 
   char tmpLine[MAX_LINE_LENGTH];
-  BaseString::snprintf(tmpLine, MAX_LINE_LENGTH, "EXTERNAL SYSTEM_");
-  strncat(tmpLine, system, MAX_LINE_LENGTH);
-  strncat(tmpLine, ":NoOfConnections", MAX_LINE_LENGTH);
+  BaseString::snprintf(tmpLine, MAX_LINE_LENGTH,
+                       "EXTERNAL SYSTEM_%s:NoOfConnections", system);
   ctx.m_config->put(tmpLine, nExtConnections);
-  
-  Config * ret = new Config();
-  ret->m_configValues = (struct ndb_mgm_configuration*)ctx.m_configValues.getConfigValues();
-  ret->m_oldConfig = ctx.m_config; ctx.m_config = 0;
-  return ret;
+
+  return new Config(ctx.m_configValues.getConfigValues());
 }
 
 //****************************************************************************
@@ -260,58 +262,44 @@ bool InitConfigFileParser::parseNameValuePair(Context& ctx, const char* line)
       tmp_string_split[i].trim("\r\n \t"); 
   }
 
-  // *************************************
-  // First in split is fname
-  // *************************************
+  return storeNameValuePair(ctx,
+                            tmp_string_split[0].c_str(), // fname
+                            tmp_string_split[1].c_str()); // value
+}
 
-  const char *fname= tmp_string_split[0].c_str();
 
-  if (!ctx.m_currentInfo->contains(fname)) {
+bool
+InitConfigFileParser::storeNameValuePair(Context& ctx,
+					 const char* fname,
+					 const char* value)
+{
+
+  if (ctx.m_currentSection->contains(fname))
+  {
+    ctx.reportError("[%s] Parameter %s specified twice", ctx.fname, fname);
+    return false;
+  }
+
+  if (!ctx.m_currentInfo->contains(fname))
+  {
     ctx.reportError("[%s] Unknown parameter: %s", ctx.fname, fname);
     return false;
   }
+
   ConfigInfo::Status status = m_info->getStatus(ctx.m_currentInfo, fname);
   if (status == ConfigInfo::CI_NOTIMPLEMENTED) {
     ctx.reportWarning("[%s] %s not yet implemented", ctx.fname, fname);
   }
-  if (status == ConfigInfo::CI_DEPRICATED) {
+  if (status == ConfigInfo::CI_DEPRECATED) {
     const char * desc = m_info->getDescription(ctx.m_currentInfo, fname);
     if(desc && desc[0]){
-      ctx.reportWarning("[%s] %s is depricated, use %s instead", 
+      ctx.reportWarning("[%s] %s is deprecated, use %s instead",
 			ctx.fname, fname, desc);
     } else if (desc == 0){
-      ctx.reportWarning("[%s] %s is depricated", ctx.fname, fname);
-    } 
+      ctx.reportWarning("[%s] %s is deprecated", ctx.fname, fname);
+    }
   }
 
-  // ***********************
-  //  Store name-value pair
-  // ***********************
-
-  return storeNameValuePair(ctx, fname, tmp_string_split[1].c_str());
-}
-
-
-//****************************************************************************
-//  STORE NAME-VALUE pair in properties section 
-//****************************************************************************
-
-bool 
-InitConfigFileParser::storeNameValuePair(Context& ctx,
-					 const char* fname, 
-					 const char* value) {
-  
-  const char * pname = fname;
-
-  if (ctx.m_currentSection->contains(pname)) {
-    ctx.reportError("[%s] Parameter %s specified twice", ctx.fname, fname);
-    return false;
-  }
-  
-  // ***********************
-  //  Store name-value pair
-  // ***********************
-
   const ConfigInfo::Type type = m_info->getType(ctx.m_currentInfo, fname);
   switch(type){
   case ConfigInfo::CI_BOOL: {
@@ -320,7 +308,7 @@ InitConfigFileParser::storeNameValuePair(Context& ctx,
       ctx.reportError("Illegal boolean value for parameter %s", fname);
       return false;
     }
-    MGM_REQUIRE(ctx.m_currentSection->put(pname, value_bool));
+    require(ctx.m_currentSection->put(fname, value_bool));
     break;
   }
   case ConfigInfo::CI_INT:
@@ -332,21 +320,70 @@ InitConfigFileParser::storeNameValuePair(Context& ctx,
     }
     if (!m_info->verify(ctx.m_currentInfo, fname, value_int)) {
       ctx.reportError("Illegal value %s for parameter %s.\n"
-		      "Legal values are between %Lu and %Lu", value, fname,
-		      m_info->getMin(ctx.m_currentInfo, fname), 
-		      m_info->getMax(ctx.m_currentInfo, fname));
+                      "Legal values are between %llu and %llu", value, fname,
+                      m_info->getMin(ctx.m_currentInfo, fname),
+                      m_info->getMax(ctx.m_currentInfo, fname));
       return false;
     }
     if(type == ConfigInfo::CI_INT){
-      MGM_REQUIRE(ctx.m_currentSection->put(pname, (Uint32)value_int));
+      require(ctx.m_currentSection->put(fname, (Uint32)value_int));
     } else {
-      MGM_REQUIRE(ctx.m_currentSection->put64(pname, value_int));
+      require(ctx.m_currentSection->put64(fname, value_int));
     }
     break;
   }
   case ConfigInfo::CI_STRING:
-    MGM_REQUIRE(ctx.m_currentSection->put(pname, value));
+    require(ctx.m_currentSection->put(fname, value));
     break;
+
+  case ConfigInfo::CI_ENUM:{
+    Uint32 value_int;
+    if (!m_info->verify_enum(ctx.m_currentInfo, fname, value, value_int)) {
+      BaseString values;
+      m_info->get_enum_values(ctx.m_currentInfo, fname, values);
+      ctx.reportError("Illegal value '%s' for parameter %s. "
+		      "Legal values are: '%s'", value, fname,
+                      values.c_str());
+      return false;
+    }
+    require(ctx.m_currentSection->put(fname, value_int));
+    break;
+  }
+
+  case ConfigInfo::CI_BITMASK:{
+    if (strlen(value) <= 0)
+    {
+      ctx.reportError("Illegal value '%s' for parameter %s. "
+                      "Error: Zero length string",
+                      value, fname);
+      return false;
+    }
+    Uint64 max = m_info->getMax(ctx.m_currentInfo, fname);
+    SparseBitmask mask((unsigned)max);
+    int res = parse_mask(value, mask);
+    if (res < 0)
+    {
+      BaseString desc("Unknown error.");
+      switch(res)
+      {
+      case -1:
+        desc.assign("Invalid syntax for bitmask");
+        break;
+      case -2:
+        desc.assfmt("Too large id used in bitmask, max is %llu", max);
+        break;
+      default:
+        break;
+      }
+
+      ctx.reportError("Illegal value '%s' for parameter %s. Error: %s",
+                      value, fname, desc.c_str());
+      return false;
+    }
+    require(ctx.m_currentSection->put(fname, value));
+    break;
+  }
+
   case ConfigInfo::CI_SECTION:
     abort();
   }
@@ -441,7 +478,7 @@ bool InitConfigFileParser::convertStringToBool(const char* s, bool& val) {
 //****************************************************************************
 static void
 trim(char * str){
-  int len = strlen(str);
+  int len = (int)strlen(str);
   for(len--;
       (str[len] == '\r' || str[len] == '\n' || 
        str[len] == ' ' || str[len] == '\t') && 
@@ -544,16 +581,18 @@ bool
 InitConfigFileParser::storeSection(Context& ctx){
   if(ctx.m_currentSection == NULL)
     return true;
-  for(int i = strlen(ctx.fname) - 1; i>=0; i--){
+  for(int i = (int)strlen(ctx.fname) - 1; i>=0; i--){
     ctx.fname[i] = toupper(ctx.fname[i]);
   }
-  BaseString::snprintf(ctx.pname, sizeof(ctx.pname), ctx.fname);
+  BaseString::snprintf(ctx.pname, sizeof(ctx.pname), "%s", ctx.fname);
+
   char buf[255];
   if(ctx.type == InitConfigFileParser::Section)
     BaseString::snprintf(buf, sizeof(buf), "%s", ctx.fname);
   if(ctx.type == InitConfigFileParser::DefaultSection)
     BaseString::snprintf(buf, sizeof(buf), "%s DEFAULT", ctx.fname);
-  BaseString::snprintf(ctx.fname, sizeof(ctx.fname), buf);
+  BaseString::snprintf(ctx.fname, sizeof(ctx.fname), "%s", buf);
+
   if(ctx.type == InitConfigFileParser::Section){
     for(int i = 0; i<m_info->m_NoOfRules; i++){
       const ConfigInfo::SectionRule & rule = m_info->m_SectionRules[i];
@@ -585,8 +624,8 @@ InitConfigFileParser::Context::reportError(const char * fmt, ...){
   if (fmt != 0)
     BaseString::vsnprintf(buf, sizeof(buf)-1, fmt, ap);
   va_end(ap);
-  fprintf(m_errstream, "Error line %d: %s\n",
-	  m_lineno, buf);
+  g_eventLogger->error("at line %d: %s",
+                       m_lineno, buf);
 
   //m_currentSection->print();
 }
@@ -600,12 +639,15 @@ InitConfigFileParser::Context::reportWarning(const char * fmt, ...){
   if (fmt != 0)
     BaseString::vsnprintf(buf, sizeof(buf)-1, fmt, ap);
   va_end(ap);
-  fprintf(m_errstream, "Warning line %d: %s\n",
-	  m_lineno, buf);
+  g_eventLogger->warning("at line %d: %s",
+                         m_lineno, buf);
 }
 
 #include <my_sys.h>
 #include <my_getopt.h>
+#ifdef HAVE_MY_DEFAULT_H
+#include <my_default.h>
+#endif
 
 static int order = 1;
 static 
@@ -627,50 +669,45 @@ InitConfigFileParser::store_in_properties(Vector<struct my_option>& options,
 {
   for(unsigned i = 0; i<options.size(); i++)
   {
-    if(options[i].comment && 
-       options[i].app_type && 
-       strcmp(options[i].comment, name) == 0)
+    if (options[i].app_type == 0)
     {
-      Uint64 value_int;
+      // Option not found in in my.cnf
+      continue;
+    }
+
+    const char* section = options[i].comment;
+    if (!section)
+    {
+      // Option which is not to be saved, like "ndbd", "ndbapi", "mysqld" etc.
+      continue;
+    }
+
+    if (strcmp(section, name) == 0)
+    {
+      const char* value = NULL;
+      char buf[32];
       switch(options[i].var_type){
       case GET_INT:
-	value_int = *(Uint32*)options[i].value;
+      case GET_UINT:
+        BaseString::snprintf(buf, sizeof(buf), "%u",
+                             *(Uint32*)options[i].value);
+        value = buf;
 	break;
-      case GET_LL:
-	value_int = *(Uint64*)options[i].value;
+      case GET_ULL:
+        BaseString::snprintf(buf, sizeof(buf), "%llu",
+                             *(Uint64*)options[i].value);
+        value = buf;
 	break;
       case GET_STR:
-	ctx.m_currentSection->put(options[i].name, *(char**)options[i].value);
-	continue;
+        value = *(char**)options[i].value;
+        break;
       default:
 	abort();
       }
 
-      const char * fname = options[i].name;
-      if (!m_info->verify(ctx.m_currentInfo, fname, value_int)) {
-	ctx.reportError("Illegal value %lld for parameter %s.\n"
-			"Legal values are between %Lu and %Lu", 
-			value_int, fname,
-			m_info->getMin(ctx.m_currentInfo, fname), 
-			m_info->getMax(ctx.m_currentInfo, fname));
-	return false;
-      }
-
-      ConfigInfo::Status status = m_info->getStatus(ctx.m_currentInfo, fname);
-      if (status == ConfigInfo::CI_DEPRICATED) {
-	const char * desc = m_info->getDescription(ctx.m_currentInfo, fname);
-	if(desc && desc[0]){
-	  ctx.reportWarning("[%s] %s is depricated, use %s instead", 
-			    ctx.fname, fname, desc);
-	} else if (desc == 0){
-	  ctx.reportWarning("[%s] %s is depricated", ctx.fname, fname);
-	} 
-      }
-      
-      if (options[i].var_type == GET_INT)
-	ctx.m_currentSection->put(options[i].name, (Uint32)value_int);
-      else
-	ctx.m_currentSection->put64(options[i].name, value_int);	
+      const char* fname = options[i].name;
+      if (!storeNameValuePair(ctx, fname, value))
+        return false;
     }
   }
   return true;
@@ -703,6 +740,9 @@ load_defaults(Vector<struct my_option>& options, const char* groups[])
   BaseString group_suffix;
 
   const char *save_file = my_defaults_file;
+#if MYSQL_VERSION_ID >= 50508
+  const
+#endif
   char *save_extra_file = my_defaults_extra_file;
   const char *save_group_suffix = my_defaults_group_suffix;
 
@@ -727,7 +767,7 @@ load_defaults(Vector<struct my_option>& options, const char* groups[])
 
   char ** tmp = (char**)argv;
   int ret = load_defaults("my", groups, &argc, &tmp);
-  
+
   my_defaults_file = save_file;
   my_defaults_extra_file = save_extra_file;
   my_defaults_group_suffix = save_group_suffix;
@@ -785,19 +825,21 @@ InitConfigFileParser::parse_mycnf()
 	opt.var_type = GET_INT;
 	break;
       case ConfigInfo::CI_INT: 
-	opt.value = (uchar**)malloc(sizeof(int));
-	opt.var_type = GET_INT;
+	opt.value = (uchar**)malloc(sizeof(uint));
+	opt.var_type = GET_UINT;
 	break;
       case ConfigInfo::CI_INT64:
-	opt.value = (uchar**)malloc(sizeof(Int64));
-	opt.var_type = GET_LL;
+	opt.value = (uchar**)malloc(sizeof(Uint64));
+	opt.var_type = GET_ULL;
 	break;
+      case ConfigInfo::CI_ENUM:
       case ConfigInfo::CI_STRING: 
+      case ConfigInfo::CI_BITMASK:
 	opt.value = (uchar**)malloc(sizeof(char *));
 	opt.var_type = GET_STR;
 	break;
       default:
-	continue;
+        continue;
       }
       opt.name = param._fname;
       opt.id = 256;
@@ -854,7 +896,7 @@ InitConfigFileParser::parse_mycnf()
     api = &options[idx+3];
   }
   
-  Context ctx(m_info, m_errstream); 
+  Context ctx(m_info);
   const char *groups[]= { "cluster_config", 0 };
   if (load_defaults(options, groups))
     goto end;
@@ -910,6 +952,9 @@ InitConfigFileParser::parse_mycnf()
 	const char * defaults_groups[] = { 0,  0, 0 };
 	for(unsigned j = 0; j<list.size(); j++)
 	{
+          // Remove leading and trailing spaces from hostname
+          list[j].trim();
+
 	  BaseString group_idx;
 	  BaseString group_host;
 	  group_idx.assfmt("%s.%s.%d", groups[0], 
@@ -926,11 +971,42 @@ InitConfigFileParser::parse_mycnf()
 	  ctx.m_userDefaults = getSection(ctx.fname, ctx.m_defaults);
 	  require((ctx.m_currentInfo = m_info->getInfo(ctx.fname)) != 0);
 	  require((ctx.m_systemDefaults = m_info->getDefaults(ctx.fname))!= 0);
-	  ctx.m_currentSection->put("HostName", list[j].c_str());
 	  if(!load_mycnf_groups(options, ctx, sections[i].name, 
 				defaults_groups))
 	    goto end;
-	  
+
+          // The [cluster_config] section in my.cnf specifies the hostname,
+          // but it can also be specified a second time in the nodes section
+          // make sure they match if specified in both places, else
+          // save the value from cluster_config section
+          //
+          // Example:
+          // [cluster_config]
+          // ndbd=hostname1
+          //      ^^^^^^^^^
+          // [cluster_config.ndbd.1]
+          // HostName=hostname1
+          //          ^^^^^^^^^
+          //
+          if (ctx.m_currentSection->contains("HostName"))
+          {
+            // HostName specified a second time, check that it matches
+            const char* host_name;
+            require(ctx.m_currentSection->get("HostName", &host_name));
+            if (strcmp(host_name, list[j].c_str()))
+            {
+              ctx.reportError("Illegal value 'HostName=%s' specified for "
+                              "%s, previously set to '%s'",
+                              host_name, group_idx.c_str(),
+                              list[j].c_str());
+              goto end;
+            }
+          }
+          else
+          {
+            require(ctx.m_currentSection->put("HostName", list[j].c_str()));
+          }
+
 	  if(!storeSection(ctx))
 	    goto end;
 	}
