@@ -1,5 +1,5 @@
 /*
-   Copyright (c) 2004, 2010, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2004, 2011, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -28,18 +28,36 @@ extern NdbMutex *g_ndb_connection_mutex;
 
 class TransporterFacade;
 class ConfigRetriever;
-class NdbThread;
-class ndb_mgm_configuration;
+struct NdbThread;
+struct ndb_mgm_configuration;
 class Ndb;
 
 extern "C" {
   void* run_ndb_cluster_connection_connect_thread(void*);
 }
 
+struct NdbApiConfig
+{
+  NdbApiConfig() :
+    m_scan_batch_size(MAX_SCAN_BATCH_SIZE),
+    m_batch_byte_size(SCAN_BATCH_SIZE),
+    m_batch_size(DEF_BATCH_SIZE),
+    m_waitfor_timeout(120000),
+    m_default_queue_option(0)
+    {}
+
+  Uint32 m_scan_batch_size;
+  Uint32 m_batch_byte_size;
+  Uint32 m_batch_size;
+  Uint32 m_waitfor_timeout; // in milli seconds...
+  Uint32 m_default_queue_option;
+};
+
 class Ndb_cluster_connection_impl : public Ndb_cluster_connection
 {
   Ndb_cluster_connection_impl(const char *connectstring,
-                              Ndb_cluster_connection *main_connection);
+                              Ndb_cluster_connection *main_connection,
+                              int force_api_nodeid);
   ~Ndb_cluster_connection_impl();
 
   void do_test();
@@ -49,15 +67,18 @@ class Ndb_cluster_connection_impl : public Ndb_cluster_connection
   Uint32 get_next_alive_node(Ndb_cluster_connection_node_iter &iter);
 
   inline unsigned get_connect_count() const;
+  inline unsigned get_min_db_version() const;
 public:
   inline Uint64 *get_latest_trans_gci() { return &m_latest_trans_gci; }
 
 private:
   friend class Ndb;
   friend class NdbImpl;
+  friend class NdbWaitGroup;
   friend void* run_ndb_cluster_connection_connect_thread(void*);
   friend class Ndb_cluster_connection;
   friend class NdbEventBuffer;
+  friend class SignalSender;
   
   struct Node
   {
@@ -73,9 +94,15 @@ private:
 
   Vector<Node> m_all_nodes;
   int init_nodes_vector(Uint32 nodeid, const ndb_mgm_configuration &config);
+  int configure(Uint32 nodeid, const ndb_mgm_configuration &config);
   void connect_thread();
   void set_name(const char *name);
-  
+  Uint32 get_db_nodes(Uint8 nodesarray[MAX_NDB_NODES]) const;
+
+  int connect(int no_retries,
+              int retry_delay_in_seconds,
+              int verbose);
+
   Ndb_cluster_connection *m_main_connection;
   GlobalDictCache *m_globalDictCache;
   TransporterFacade *m_transporter_facade;
@@ -84,7 +111,6 @@ private:
   int (*m_connect_callback)(void);
 
   int m_optimized_node_selection;
-  char *m_name;
   int m_run_connect_thread;
   NdbMutex *m_event_add_drop_mutex;
   Uint64 m_latest_trans_gci;
@@ -96,6 +122,18 @@ private:
 
   BaseString m_latest_error_msg;
   unsigned m_latest_error;
+
+  // Scan batch configuration parameters
+  NdbApiConfig m_config;
+  
+  // keep initial transId's increasing...
+  Uint32 m_max_trans_id;
+
+  // Base offset for stats, from Ndb objects that are no 
+  // longer with us
+  Uint64 globalApiStatsBaseline[ Ndb::NumClientStatistics ];
+
+  NdbWaitGroup *m_multi_wait_group;
 };
 
 #endif
