@@ -1,5 +1,5 @@
 /*
-   Copyright (c) 2003, 2005, 2006 MySQL AB, 2009 Sun Microsystems, Inc.
+   Copyright (c) 2005, 2006, 2008 MySQL AB, 2008, 2009 Sun Microsystems, Inc.
    Use is subject to license terms.
 
    This program is free software; you can redistribute it and/or modify
@@ -107,7 +107,7 @@ double userGetTimeSync(void)
 int 
 userDbCommit(UserHandle *uh){
   if(uh->pCurrTrans != 0){
-    int check = uh->pCurrTrans->execute( Commit ); 
+    /* int check = */ uh->pCurrTrans->execute( Commit ); 
     NdbError err = uh->pCurrTrans->getNdbError();
     uh->pNDB->closeTransaction(uh->pCurrTrans);
     uh->pCurrTrans = 0;
@@ -129,15 +129,7 @@ userDbCommit(UserHandle *uh){
  * TRUE - Normal table
  * FALSE - Table w.o. checkpoing and logging
  */
-
-#ifdef	__cplusplus
-extern "C" {
-#endif
 extern int useTableLogging;
-#ifdef	__cplusplus
-}
-#endif
-
 
 int
 create_table_server(Ndb * pNdb){
@@ -287,7 +279,7 @@ create_table_group(Ndb * pNdb){
     error_handler("createAttribute (group id)", 
 		  MySchemaTransaction->getNdbError(), 0);
 
-  check = MySchemaOp->createAttribute( GROUP_NAME,
+  check = MySchemaOp->createAttribute( NDB_GROUP_NAME,
 				       NoKey,
 				       sizeof(char) << 3,
 				       GROUP_NAME_LENGTH,
@@ -658,8 +650,10 @@ int userDbInsertSubscriber(UserHandle      *uh,
   int check;
   uint32 activeSessions = 0;
   Location l = 0;
-  ChangedBy changedBy; snprintf(changedBy, sizeof(changedBy), "ChangedBy");
-  ChangedTime changedTime; snprintf(changedTime, sizeof(changedTime), "ChangedTime"); 
+  ChangedBy changedBy;
+  ChangedTime changedTime;
+  BaseString::snprintf(changedBy, sizeof(changedBy), "ChangedBy");
+  BaseString::snprintf(changedTime, sizeof(changedTime), "ChangedTime");
 
   NdbConnection * MyTransaction = 0;
   if(uh->pCurrTrans != 0){
@@ -727,7 +721,7 @@ int userDbInsertGroup(UserHandle *uh,
   check = MyOperation->equal(GROUP_ID, (char*)&groupId);
   CHECK_MINUS_ONE(check, "equal", MyTransaction);  
   
-  check = MyOperation->setValue(GROUP_NAME, name);
+  check = MyOperation->setValue(NDB_GROUP_NAME, name);
   CHECK_MINUS_ONE(check, "setValue name", MyTransaction);  
 
   check = MyOperation->setValue(GROUP_ALLOW_READ, (char*)&allowRead);
